from heapq import heappush, heappop
import random
import matplotlib.pyplot as plt
from timeit import timeit
import math
import copy
import itertools

def SP_naive (graph, s):
    '''
    Shortest path algorithm with naive implementation.
    graph: adjacency list of the graph
    s: source vertex
    return: a dictionary of shortest distances from s to all other vertices
    '''
    frontier = [s]
    dist = {s:0}

    while len(frontier) > 0:

        x = min(frontier, key = lambda k: dist[k])
        frontier.remove(x)

        for y, dxy in graph[x].items():
            dy = dist[x] + dxy

            if y not in dist:
                frontier.append(y)
                dist[y] = dy

            elif dist[y] > dy:
                dist[y] = dy

    return dist

# Shortest path algorithm with binary heap

def SP_heap (graph, s):
    '''
    Shortest path algorithm with binary heap.
    graph: adjacency list of the graph
    s: source vertex
    return: a dictionary of shortest distances from s to all other vertices
    '''
    frontier = []
    heappush(frontier, (0, s))
    done = set()
    dist = {s: 0}

    while len(frontier) > 0:

        dx, x = heappop(frontier)
        if x in done:
            continue

        done.add(x)

        for y, dxy in graph[x].items():
            dy = dx + dxy

            if y not in dist or dist[y] > dy:
                heappush(frontier,(dy, y))
                dist[y] = dy

    return dist

####################### Question 1 #####################
def q1():
    return '''
    Insérer votre réponse à la question 1 ici.
    '''

####################### Question 2 #####################
def q2():
    return '''
    Insérer votre réponse à la question 2 ici.
    '''

def random_sparse_graph (n, step):
    '''
    Generate a random sparse graph with n vertices.
    n: number of vertices
    step: maximum distance between two adjacent vertices
    return: adjacency list of the graph
    '''
    graph = {f'{i}_{j}': {} for i in range(1, n+1) for j in range(1, n+1)}

    for i in range(1, n+1):
        for j in range(1, n):
            d = random.randint(step+1, 2*step)
            graph[f'{i}_{j}'][f'{i}_{j+1}'] = d
            graph[f'{i}_{j+1}'][f'{i}_{j}'] = d

    for i in range(1, n):
        for j in range(1, n+1):
            d = random.randint(step+1, 2*step)
            graph[f'{i}_{j}'][f'{i+1}_{j}'] = d
            graph[f'{i+1}_{j}'][f'{i}_{j}'] = d

    return graph

def random_dense_graph (n, d_max):
    '''
    Generate a random dense graph with n vertices.
    n: number of vertices
    d_max: maximum distance between two adjacent vertices
    return: adjacency list of the graph
    '''
    graph = {f'{i}':{} for i in range(n)}

    for n1 in graph:
        for n2 in graph:
            if n2!= n1 and n2 not in graph[n1]:
                d = random.randint(1, d_max)
                graph[n1][n2] = d
                graph[n2][n1] = d

    return graph

def benchmark():
    Time_heap_sparse = []
    Time_naive_sparse = []
    Time_heap_dense = []
    Time_naive_dense = []

    n_list = []

    for N in range(10, 30):
        print(f'N={N} is being processed...')
        n = N*N
        n_list.append(n)

        # on compare sur des cartes non-denses: grille de côté N contient N*N villes
        graph_sparse = random_sparse_graph(n = N, step = 100)

        # on calcule une moyenne sur N lancements en tirant aléatoirement une ville de départ à chaque fois
        Time_naive_sparse.append(timeit(lambda: SP_naive(graph_sparse, random.choice(list(graph_sparse))), number=N) / N)
        Time_heap_sparse.append(timeit(lambda: SP_heap(graph_sparse, random.choice(list(graph_sparse))), number=N) / N)

        # on compare sur des cartes denses
        graph_dense = random_dense_graph(n = N*N, d_max = 10000)

        Time_naive_dense.append(timeit(lambda: SP_naive(graph_dense, random.choice(list(graph_dense))), number=N) / N)
        Time_heap_dense.append(timeit(lambda: SP_heap(graph_dense, random.choice(list(graph_dense))), number=N) / N)

    plt.xlabel('N')
    plt.ylabel('T')
    plt.plot(n_list, Time_naive_sparse, 'r^', label="naive sparse")
    plt.plot(n_list, Time_heap_sparse, 'b^', label="heap sparse")
    plt.plot(n_list, Time_naive_dense, 'r*', label="naive dense")
    plt.plot(n_list, Time_heap_dense, 'b*', label="heap dense")

    plt.legend()
    plt.show()
    
# From your answer to the precedent question, what structure has the 
# better complexity for dense graphs in theory? What you observe from 
# the result of the benchmark? Can you explain this?

####################### Question 3 #####################
def q3():
    print("Benchmarking...")
    benchmark()
    return '''
    Insérer votre réponse à la question 3 ici.
    '''

# For sparse graphs, does the running result of the benchmark for the naive 
# structure (list) correspond to the theoretical time complexity? 
# Can you explain this?

def q4():
    return '''
    Insérer votre réponse à la question 4 ici.
    '''

# What is your general conclusion?
def q5():
    return '''
    Insérer votre réponse à la question 5 ici.
    '''

####################### Question 6 #####################

def add_source(graph, src):
    '''
    Add a source vertex to the graph.
    graph: adjacency list of the graph
    src: source vertex
    return: adjacency list of the graph with the source vertex
    '''
    if src in graph:
        return "error: the source vertex is already in the graph"

    graph_src=graph.copy()

    ############TODO : complete code#########

    return graph_src

####################### Question 7 #####################

def bellman_ford(graph, src):
    '''
    Bellman-Ford algorithm to find the shortest path from a source vertex to all other vertices.
    graph: adjacency list of the graph
    src: source vertex
    return: a dictionary of shortest distances from the source vertex to all other vertices
    '''
    n=len(graph)

    dist={}

    #initialize dist
    ############TODO : complete code#############
    for i in graph:
        dist[i]=math.inf
    dist[src]=0

    # calculate optimal distance
    ############TODO : complete code#############

    # detect negative cycle: return None if there is any
    ############TODO : complete code#############

    return dist


####################### Question 8 #####################

def rewrite_weights(graph, dist):
    '''
    Rewrite the weights of the graph to make them nonnegative.
    graph: adjacency list of the graph
    dist: a dictionary of shortest distances from the source vertex to all other vertices
    return: adjacency list of the graph with nonnegative weights
    '''
    # use deepcopy
    altered_graph = copy.deepcopy(graph)

    # Recalculate the new nonnegative weights
    ############### TODO : complete code ##################

    return altered_graph

####################### Question 9 #####################
def q9():
    return '''
    Insérer votre réponse à la question 9 ici.
    '''

####################### Question 10 #####################
def q10():
    return '''
    Insérer votre réponse à la question 10 ici.
    '''

####################### Question 11 #####################

def all_distances(graph):
    '''
    Compute all shortest distances between all pairs of vertices in the graph.
    graph: adjacency list of the graph
    return: a dictionary of all shortest distances between all pairs of vertices
    '''
    d = {(u,v):None for u in graph for v in graph}

    ############### TODO : complete code ##################

    return d

####################### Question 12 #####################

def BF_SP_all_pairs(graph, src="source"):

    ############### TODO : complete code ##################
    d = {(u,v):None for u in graph for v in graph}

    # Return a dictionary of distances for all pairs of nodes

    return d

####################### Question 13 #####################

def q13():
    return '''
    Insérer votre réponse à la question 13 ici.
    '''

####################### Question 14 #####################

def closest_oven(house, oven_houses, distance_dict):
    '''
    Find the closest oven house to a given house.
    house: a house
    oven_houses: a list of oven houses
    distance_dict: a dictionary of all shortest distances between all pairs of vertices
    return: a tuple of the distance to the closest oven house and the closest oven house
    '''
    ############### TODO : complete code ##################
    return (0, None)

####################### Question 15 #####################

def kcentre_value(village, oven_houses, distance_dict):
    '''
    Compute the maximum distance between a house and the closest oven house.
    village: adjacency list of the village
    oven_houses: a list of oven houses
    distance_dict: a dictionary of all shortest distances between all pairs of vertices
    return: the maximum distance between a house and the closest oven house
    '''

    ############### TODO : complete code ##################
    return 0
    
####################### Question 16 #####################

def read_map(filename):
    '''
    Read a map from a file.
    filename: name of the file
    return: adjacency list of the map
    '''
    f = open(file=filename, mode='r', encoding='utf-8')

    map = {}
    while True:  # reading list of cities from file
        ligne = f.readline().rstrip()
        if (ligne == '--'):
            break
        info = ligne.split(':')
        map[info[0]] = {}

    while True:  # reading list of distances from file
        ligne = f.readline().rstrip()
        if (ligne == ''):
            break
        info = ligne.split(':')
        map[info[0]][info[1]] = int(info[2])

    return map

def brute_force(map, candidates, k, distance_dict) :
    best_combi = []
    best_dist = math.inf
    for combi in itertools.combinations(candidates, k):
        dist= kcentre_value(map, list(combi), distance_dict)
        if  dist<best_dist:
            best_combi= list(combi)
            best_dist= dist
    return  best_dist, set(best_combi)

def BF_benchmark():

    village = read_map('village.map')

    village_distance_dict = BF_SP_all_pairs(village)

    # assert brute_force(village, list(village), 3, village_distance_dict) == (0, {'C', 'D', 'B'})
    # assert brute_force(village, list(village), 2, village_distance_dict) == (8, {'C', 'A'})

    Time_brute_force = []

    k_list = []

    for k in range(1, 20):
        print(f'k={k} is being processed...')
        Time_brute_force.append(timeit(lambda: brute_force(village, list(village), k, village_distance_dict), number=1))
        k_list.append(k)

    #print N_list, time_list
    plt.xlabel('k')
    plt.ylabel('T')
    plt.plot(k_list, Time_brute_force, 'r^')
    plt.xticks(k_list)
    plt.show()
    
def q16():
    print("Benchmarking...")
    BF_benchmark()
    return'''
    Insérer votre réponse à la question 16 ici.
    '''
####################### Question 17 #####################

def greedy_algorithm(map, candidates, k, distance_dict):
    '''
    Greedy algorithm to find the k-centre of a village.
    map: adjacency list of the village
    candidates: list of houses in the village
    k: number of oven houses
    distance_dict: a dictionary of all shortest distances between all pairs of vertices
    return: a tuple of the maximum distance between a house and the closest oven house and the set of oven houses
    '''
    ############### TODO : complete code ##################
    return 0, set()
        
####################### Question 18 #####################

def random_algorithm(map, candidates, k, distance_dict, trials=100) :
    '''
    Random algorithm to find the k-centre of a village.
    map: adjacency list of the village
    candidates: list of houses in the village
    k: number of oven houses
    distance_dict: a dictionary of all shortest distances between all pairs of vertices
    trials: number of trials
    return: a tuple of the maximum distance between a house and the closest oven house and the set of oven houses
    '''
    ############### TODO : complete code ##################
    return 0, set()

####################### Question 19 #####################

def BF_G_R_benchmark(max_k, random_trials = 100):
    '''
    Benchmark the brute force, greedy and random algorithms.
    max_k: maximum number of oven houses
    random_trials: number of trials for the random algorithm
    '''
    village = read_map('village.map')

    village_distance_dict = BF_SP_all_pairs(village)

    # compare on lite
    time_bruteforce_lite = []
    time_random_lite = []
    time_greedy_lite = []
 

    d_bruteforce_lite = []
    d_random_lite = []
    d_greedy_lite = []
 

    k_list = []


    for k in range(2, max_k):
        print(f'k={k} is being processed...')
        k_list.append(k)


        time_bruteforce_lite.append(timeit(lambda: d_bruteforce_lite.append(brute_force(village, list(village), k, village_distance_dict)[0]), number=1))

        time_greedy_lite.append(timeit(lambda: d_greedy_lite.append(greedy_algorithm(village, list(village), k, village_distance_dict)[0]), number=1))
        time_random_lite.append(timeit(lambda: d_random_lite.append(random_algorithm(village, list(village), k, village_distance_dict, random_trials)[0]), number=1))


    plt.subplot(2, 1, 1)
    plt.xlabel('k')
    plt.ylabel('T')
    plt.xticks(k_list)
    plt.plot(k_list, time_bruteforce_lite, 'r*')
    plt.plot(k_list, time_greedy_lite, 'b*')
    plt.plot(k_list, time_random_lite, 'y*')

    plt.subplot(2, 1, 2)
    plt.xlabel('k')
    plt.ylabel('d')
    plt.xticks(k_list)
    plt.plot(k_list, d_bruteforce_lite, 'r-')
    plt.plot(k_list, d_greedy_lite, 'b-')
    plt.plot(k_list, d_random_lite, 'y-')
    plt.show()
    
def q19():
    print("Benchmarking...")
    BF_G_R_benchmark(10)
    return '''
    Insérer votre réponse à la question 19 ici.
    '''